import argparse
import base64
import boto3
import json
import logging
import os
import time
import subprocess

from awspot.managers import ec2Manager

def parse_args():
    # TODO: Break ec2 commands into separate scripts w/separate args
    parser = argparse.ArgumentParser(description='Script to manage ec2 spot instances.')
    parser.add_argument('instance_type', type=str,
                        choices=['ec2'],
                        help='request_type (ec2 or spot fleet)')
    parser.add_argument('command', type=str,
                        choices=['launch', 'list_running', 'terminate',
                                 'ssh', 'jupytunnel'],
                        help='action type')
    parser.add_argument('-n', '--name', type=str,
                        help='name for instance')
    parser.add_argument('-s', '--specification', type=str,
                        help='path to specification JSON file')
    parser.add_argument('-d', '--userdata', type=str, default=None,
                        help='optional path to userdata shell script')
    parser.add_argument('-p', '--price', type=str,
                        help='max bid price for instance')
    parser.add_argument('-k', '--key_file', type=str,
                        help='path to .pem file')
    parser.add_argument('-u', '--user', type=str,
                        help='user to login as')
    parser.add_argument('-P', '--profile', type=str,
                        help='AWS profile to use', default='default')

    return parser.parse_args()


#################################################
# SCRIPT                                        #
#################################################
args = parse_args()

session = boto3.Session(profile_name=args.profile)
client = session.client('ec2')
manager = ec2Manager(client)

if args.command == 'launch':
    # TODO: Add console logging.
    manager.launch_instance(args.name,
                            args.specification,
                            args.userdata,
                            args.price)

elif args.command == 'list_running':
    manager.list_running_instances()

elif args.command == 'ssh':
    # TODO: Add key pair management
    instance = manager.find_instance_by_name(args.name)
    public_dns = instance['PublicDnsName']
    key_file = args.key_file
    user_name = args.user

    subprocess.run(["ssh", "-i", key_file,
                    f"{user_name}@{public_dns}"])

elif args.command == 'jupytunnel':
    instance = manager.find_instance_by_name(args.name)
    public_dns = instance['PublicDnsName']
    key_file = args.key_file
    user_name = args.user

    subprocess.run(["ssh", "-i", key_file,
                    "-N", "-L", "8000:localhost:8888",
                    f"{user_name}@{public_dns}"])

elif args.command == 'terminate':
    # TODO: Add console logging.
    manager.terminate(args.name)



import json

class Manager(object):
    """ Base class for spot resource managers."""

    def __init__(self, client):
        self.client = client

    def launch(self, **kwargs):
        """ Launch resource. """
        raise NotImplementedError

    def list_running(self, **kwargs):
        """ List running instances of resource. """
        raise NotImplementedError

    def terminate(self, **kwargs):
        """ Terminate resource by name. """
        raise NotImplementedError



import base64
import json
import logging
import os
import time
import typing

from typing import List

from .base import Manager

class ec2Manager(Manager):
    """ Class for managing ec2 spot instances. """

    def launch_instance(self, name: str, launch_spec_file: str,
                        userdata_file: str, price: str):
        """ Launch ec2 spot instance, store information by name. """
        # load launch specifications and base64 encoded userdata
        with open(launch_spec_file) as f:
            launch_spec = json.loads(f.read())
        if userdata_file is not None:
            with open(userdata_file) as f:
                userdata_str = f.read()
                userdata_bytes = base64.b64encode(bytes(userdata_str, 'utf-8'))
                userdata = userdata_bytes.decode('utf-8')
                launch_spec['UserData'] = userdata

        # request instance
        request = self.client.request_spot_instances(
            LaunchSpecification=launch_spec,
            SpotPrice=price
        )
        request_id = request['SpotInstanceRequests'][0].get('SpotInstanceRequestId')

        if request_id is None:
            print('Request submission failed.')
            return False
        else:
            print(f"\nSpot instance request submitted.\nSpotInstanceRequestId: {request_id}\n")

        # check request state every 1.5 seconds, get instance id when fulfilled
        holding = ['pending-evaluation','not-scheduled-yet',
                   'pending-fulfillment']
        request_status = 'pending-evaluation'
        while request_status in holding:
            request_state = self.client.describe_spot_instance_requests(
                SpotInstanceRequestIds=[request_id]
            )['SpotInstanceRequests'][0]

            request_status = request_state['Status'].get('Code')
            time.sleep(1.5)

        # handle successful fulfillment
        if request_status == 'fulfilled':
            instance_id = request_state.get('InstanceId')
            print(f"Spot instance request fulfilled.\nInstanceId: {instance_id}\n")
            # Add name tag to resource
            self.client.create_tags(
                Resources=[instance_id],
                Tags=[{'Key': 'Name',
                       'Value': name}]
            )

            return True

        # handle failed fulfillment
        else:
            self.client.cancel_spot_instance_requests(
                SpotInstanceRequestIds=[request_id]
            )
            print(f"Spot instance request cancelled.\nReason: {request_status}\n")

            return False

    def list_running_instances(self):
        """ List active instances of resource. """
        instances = self._get_running_instances()
        if not instances:
            print("No running instances.")
        else:
            output = "\nInstanceId\t\tName\n\n"
            for instance in instances:
                name = self._get_instance_name(instance)
                instance_id = instance['InstanceId']
                output += f"{instance_id}\t{name}\n"

            print(output)

    def find_instance_by_name(self, name: str):
        """ Lookup resource by name """
        # TODO: Handle case where multiple instances have same name
        instances = self._get_running_instances()
        for instance in instances:
            if self._get_instance_name(instance) == name:
                return instance

    def terminate(self, name: str):
        """ Terminate resource by name. """
        # TODO: Create fail-safe termination check.
        instance = self.find_instance_by_name(name)
        if instance:
            instance_id = instance['InstanceId']
            response = self.client.terminate_instances(InstanceIds=[instance_id])
            if response['TerminatingInstances'] and \
               response['TerminatingInstances'][0]['InstanceId'] == instance_id:
                print(f"Instance terminating.\nName: {name}\nInstanceId: {instance_id}\n")
                return True

        return False

    def _get_instance_name(self, instance: List)->str:
        """ Get instance name from instance description """
        tags = instance['Tags']
        for tag in tags:
            if tag['Key'] == 'Name':
                return tag['Value']

    def _get_running_instances(self)->List:
        """ Get list of running instances """
        filters = {'instance-lifecycle': ['spot'],
                   'instance-state-name': ['running']}
        filters_list = [{'Name': k, 'Values': v} for k, v in filters.items()]

        response = self.client.describe_instances(Filters=filters_list)
        reservations = response.get('Reservations')
        instances = [i for r in reservations for i in r.get('Instances')]

        return instances



import datetime
import logging
import typing

from typing import List

from .base import Manager

class emrManager(Manager):
    """ Class for managing emr spot fleets. """

    def launch_cluster(self, name: str, launch_spec_file: str,
                        userdata_file: str, price: str):
        """ Launch emr cluster """
        raise NotImplementedError

    def list_active_clusters(self):
        """ List active instances of resource. """
        clusters = self._get_active_clusters()
        if not clusters:
            print("No active clusters.")
        else:
            output = "\nClusterId\t\tName\n\n"
            for cluster in clusters:
                name = cluster.get('Name')
                cluster_id = cluster.get('Id')
                output += f"{cluster_id}\t{name}\n"

        return output

    def find_cluster_by_name(self, name: str):
        """ Lookup cluster by name """
        # TODO: Handle case where multiple clusters have same name
        clusters = self._get_active_clusters()
        for cluster in clusters:
            if cluster['Name'] == name:
                return cluster

    def find_master_dns_by_name(self, name: str):
        """ Lookup cluster by name, return master public DNS """
        cluster = self.find_cluster_by_name(name)
        response = self.client.describe_cluster(ClusterId=cluster.get('Id'))
        dns = response['Cluster'].get('MasterPublicDnsName')

        return dns

    def terminate(self, name: str):
        """ Terminate resource by name. """
        raise NotImplementedError

    def _get_active_clusters(self)->List:
        """ Get list of active clusters created in last 30 days """
        cluster_states = ['STARTING', 'BOOTSTRAPPING', 'RUNNING', 'WAITING']
        created_after = datetime.datetime.now() + datetime.timedelta(-30)

        response = self.client.list_clusters(ClusterStates=cluster_states,
                                             CreatedAfter=created_after)

        clusters = response.get('Clusters')

        return clusters



import boto3

from unittest import TestCase
from unittest.mock import Mock

from awspot.managers import ec2Manager

class TestEc2Manager(TestCase):
    # TODO: Find isolated way to mock boto3 calls. Moto does not provide enough
    #       functionality.

    MOCK_CLIENT_ATTRS = {
        'request_spot_instances.return_value': {
            'SpotInstanceRequests': [{'SpotInstanceRequestId': 'sir-fa789d54'}]
        },
        'describe_spot_instance_requests.return_value': {
            'SpotInstanceRequests': [{'InstanceId': 'i-0578da97c8ee56262'}]
        },
        'describe_instances.return_value': {
            'Reservations': [{
                    'Instances': [{
                            'InstanceId': 'i-0578da97c8ee56262',
                            'Tags': [{'Key': 'Name', 'Value': 'test'}]
                        }]
                }]
        },
        'terminate_instances.return_value': {
            'TerminatingInstances': [{'InstanceId': 'i-0578da97c8ee56262'}]
        }
    }

    @classmethod
    def setUpClass(cls):
        mock_client = Mock()
        mock_client.configure_mock(**cls.MOCK_CLIENT_ATTRS)

        cls.manager = ec2Manager(mock_client)
        cls.launch_specs = './assets/ec2_launch_specs.json'
        cls.userdata = './assets/ec2_userdata.sh'
        cls.price = '0.070'

    def test_launch(self):
        result = self.manager.launch_instance('test',
                                              self.launch_specs,
                                              self.userdata,
                                              self.price)
        self.assertTrue(result)

    def test_terminate(self):
        result = self.manager.terminate('test')
        self.assertTrue(result)



import boto3

from unittest import TestCase
from unittest.mock import Mock

from awspot.managers import emrManager
from awspot.test import read_response

class TestEmrManager(TestCase):
    MOCK_CLIENT_ATTRS = {
        'list_clusters.return_value': {
            'Clusters': [{'Id': 'j-323SIFSG3JICT',
                          'Name': 'awspot-emr-test'}]
        },
        'describe_cluster.return_value':
        read_response('./assets/responses/emr_describe_cluster_response.json')
    }

    @classmethod
    def setUpClass(cls):
        mock_client = Mock()
        mock_client.configure_mock(**cls.MOCK_CLIENT_ATTRS)

        cls.manager = emrManager(mock_client)
        cls.price = '0.10'

    def test_launch(self):
        # TODO: Implement.
        pass

    def test_terminate(self):
        # TODO: Implement.
        pass

    def test_list_running(self):
        expected = "\nClusterId\t\tName\n\nj-323SIFSG3JICT\tawspot-emr-test\n"
        result = self.manager.list_active_clusters()

        self.assertEqual(expected, result)

    def test_find_dns_by_name(self):
        expected = 'ec2-34-211-56-169.us-west-2.compute.amazonaws.com'
        result = self.manager.find_master_dns_by_name('awspot-emr-test')

        self.assertEqual(expected, result)



import argparse
import boto3
import logging
import subprocess

from awspot.managers import emrManager

def parse_args():
    parser = argparse.ArgumentParser(description='Script to manage emr spot instance fleets.')
    parser.add_argument('instance_type', type=str,
                        choices=['emr'],
                        help='request_type')
    parser.add_argument('command', type=str,
                        choices=['launch', 'list_running', 'terminate',
                                 'ssh', 'jupytunnel'],
                        help='action type')
    parser.add_argument('-n', '--name', type=str,
                        help='name for instance')
    parser.add_argument('-c', '--configuration', type=str,
                        help='path to configuration JSON file')
    parser.add_argument('-b', '--bootstrap', type=str, default=None,
                        help='optional path to bootstrap shell script in s3')
    parser.add_argument('-p', '--price', type=str,
                        help='max bid price for instance(s)')
    parser.add_argument('-k', '--key_file', type=str,
                        help='path to .pem file')
    parser.add_argument('-u', '--user', type=str,
                        help='user to login as')
    parser.add_argument('-P', '--profile', type=str,
                        help='AWS profile to use', default='default')
    parser.add_argument('--port', type=int,
                        help='host port to connect to', default=8888)

    return parser.parse_args()


#################################################
# SCRIPT                                        #
#################################################
args = parse_args()

session = boto3.Session(profile_name=args.profile)
client = session.client('emr')
manager = emrManager(client)

if args.command == 'launch':
    raise NotImplementedError

elif args.command == 'list_active_clusters':
    output = manager.list_active_clusters()
    print(output)

elif args.command == 'ssh':
    # TODO: Add key pair management
    public_dns = manager.find_master_dns_by_name(args.name)
    key_file = args.key_file
    user_name = args.user

    subprocess.run(["ssh", "-i", key_file,
                    f"{user_name}@{public_dns}"])

elif args.command == 'jupytunnel':
    public_dns = manager.find_master_dns_by_name(args.name)
    key_file = args.key_file
    user_name = args.user

    subprocess.run(["ssh", "-i", key_file,
                    "-N", "-L", f"8000:localhost:{args.port}",
                    f"{user_name}@{public_dns}"])

elif args.command == 'terminate':
    raise NotImplementedError
